/*
  structures.c: Implements the hash table which addresses one file per line. 
  Each line contains a unique identification of the file, its lease expiration 
  time and a pointer to a doubly linked list of its blocks in the cache ordered
  by diminishing offset.
  Colisions in this hash table are treated with one additional linked list per
  colision.  Besides, there is a LRU doubly linked list of cache blocks.
*/

#include <linux/kernel.h>
#include <linux/mm.h>
#include "cache.h"
#include "structures.h"
#include "hash.h"
#include <string.h>



struct hash_table_entry cache_table[HASH_SIZE];

struct list_node *LRU = NULL, 
                 *MRU = NULL; /* most recently used block*/

unsigned long cache_current_size = 0;


void init_cache_table()
{
  int i;
  
  for(i = 0; i < HASH_SIZE; i++)
    {
      cache_table[i].hostname = NULL;
      cache_table[i].colision_next = NULL; 
    }
}



struct list_node *new_cache_block(struct hash_table_entry *file, 
				  struct list_node *predecessor)
/* 
   Allocs space for the block meta-data, inserts this block at the beginning of
   the LRU queue (will be pointed by MRU) and after predecessor in the offset 
   list. Returns a pointer to the new cell.
*/
{ 
  struct list_node *block = cache_malloc(sizeof(struct list_node));

  if(!block) return NULL;

  /* Inserts this block at the begining of the LRU list */

  if(MRU) 
   {
      block->LRU_next = MRU;
      MRU->LRU_prev = block;
      MRU = block;
      block->LRU_prev = NULL;
    }
  else /* the LRU list is empty */
    {
      MRU = LRU = block;
      block->LRU_prev = block->LRU_next = NULL;
    } 

  /* Inserts it at the proper place in the offset list. */

  if(!predecessor) /* Inserts at the beginning of the list */
    {
      block->offset_next = file->offset_list;
      file->offset_list = block;
      block->offset_prev = NULL;
      if(block->offset_next) block->offset_next->offset_prev = block;
    }
  else 
    {
      block->offset_next = predecessor->offset_next;
      if (block->offset_next) 
	block->offset_next->offset_prev = block;
      
      predecessor->offset_next = block;
      block->offset_prev = predecessor;
    }
  return block;
  
}/* new_cache_block */



void remove_cache_block_LRU_list(struct list_node *block)
/*
  Removes this block from the LRU list. (DOESN'T free its allocated space)
*/
{
  
  if(!block->LRU_prev)  /* Is this block the first item of the list ? */
    MRU = block->LRU_next;
  else 
    block->LRU_prev->LRU_next = block->LRU_next;

  if(!block->LRU_next) /* The last item of the list ? */
    LRU = block->LRU_prev;
  else 
    block->LRU_next->LRU_prev = block->LRU_prev;
    
}/* remove_cache_block_LRU_list */



void refresh_block_LRU_list(struct list_node *block)
/*
   Removes this block from its position in the LRU list and inserts it at the 
  beginning of the same list (pointed by MRU).
*/
{
  if(MRU != block) /* Do nothing if it's already pointed by MRU */
   {
     block->LRU_prev->LRU_next = block->LRU_next;
     if(!block->LRU_next) /* The last item of the list ? */
       LRU = block->LRU_prev;
     else 
       block->LRU_next->LRU_prev = block->LRU_prev;
    
     block->LRU_next = MRU; 
     MRU = block;             /* now it's pointed by MRU */
     block->LRU_prev = NULL;
     block->LRU_next->LRU_prev = block;
   }
}/* refresh_block_LRU_list */




void remove_cache_block_offset_list(struct list_node *block)
/*
  Removes this block from the offset list and frees its allocated space.
*/
{
  if(!block->offset_prev)  /* Is block the first item of the list ? */
    block->file->offset_list = block->offset_next;

  else block->offset_prev->offset_next = block->offset_next;

  if(block->offset_next)
    block->offset_next->offset_prev = block->offset_prev;

  PRINTK("Removi bloco: inode : %ld offset : %ld\n",block->file->inode_number,
	 block->offset);  
  cache_current_size -= (block->size + sizeof(struct list_node));
  
  if(!block->file->offset_list)     /* Remove this file from the cache if it */
    remove_file_cache(block->file); /* has no more blocks cached.            */
  
  kfree_s(block->data, block->size);
  kfree_s(block,sizeof(struct list_node));

}/* remove_cache_block_offset_list */




int remove_LRU_cache_block()
/*
   Removes the Least Recently Used block from the cache.
*/
{
  struct list_node *block = LRU;
  
  if(!LRU) return (0); /* there is no block to remove */

  LRU = block->LRU_prev;

  if(!LRU) MRU = NULL;
  else LRU->LRU_next  = NULL;

  /* Remove the block from the offset list, too */
  remove_cache_block_offset_list(block);
  return(1);

}/* remove_LRU_cache_block */




void clear_file_cache(struct hash_table_entry *file)
/*
  Removes all the blocks of this file from the cache and removes this
  file from the hash table. 
*/
{
  struct list_node *temp, *block = file->offset_list;
 
  if(!block) remove_file_cache(file);
  else
    while(block) /*  Scans the blocks of this file and */ 
      {          /* removes them from the cache.       */ 
	
	temp = block->offset_next;
	remove_cache_block_LRU_list(block);
	remove_cache_block_offset_list(block);
	block = temp;
      }
}/* clear_file_cache */




void remove_file_cache(struct hash_table_entry *file)
/*
  Removes a file from the hash table. It can only be called when the offset
  list for that file is empty.
*/
{
  int strsize;
  
  if(file->colision_next && file->colision_prev)
    {
      PRINTK("Hash colision for file %ld\n",file->inode_number);
      file->colision_next->colision_prev = file->colision_prev;
    }


  kfree_s(file->hostname,strsize = strlen(file->hostname)+1);
  cache_current_size -= strsize;

  if(!file->colision_prev)  /* Is file the first item of the list ? */
    file->hostname = NULL; /* Mark "free" so it can be used later */
  else 
    {
      file->colision_prev->colision_next = file->colision_next;
      PRINTK("The space occupied by the file %ld was freed\n",
	     file->inode_number);      
      kfree_s(file,sizeof(struct hash_table_entry));
      cache_current_size -= sizeof(struct hash_table_entry);
    }
}/* remove_file_cache */



void delete_file_from_cache(unsigned long inode_number, char *hostname)
/*
   Called by nfs_file_write and by nfs_proc_read to delete completely
   the cached data for a file.

   It would be better deleting the data blocks only. In the case of
   nfs_file_write, it would be better to put the just written blocks
   on the cache. 
*/
{
 struct hash_table_entry *file = find_hash_entry(inode_number, hostname);

 if(file) clear_file_cache(file);

}/* delete_file_from_cache */




struct list_node *find_cache_block(struct hash_table_entry *file, off_t offset)
/*
  Finds the block of greater offset less or equal to the parameter "offset".
  Returns the block meta-data address if it is found and NULL otherwise.
  (called by read_cache())
*/ 
{
  struct list_node *block;
  
  for(block = file->offset_list; 
      block && block->offset > offset;
      block = block->offset_next);
  
  return(block);
}



struct list_node *find_predecessor(struct hash_table_entry *file, off_t offset)
/* 
   Returns a pointer to the predecessor of the new cache block that will be 
   inserted. (This function is called by store_cache())
*/
{
  struct list_node *block;
  
  if(!(block = file->offset_list)) return NULL;
  
  while(block->offset_next && block->offset > offset)
      block = block->offset_next;
  
  if(block->offset <= offset) return block->offset_prev;
  else return block;
  
}/* find_predecessor */

  

void *cache_malloc(unsigned int size)
{
  int b = 1;
  void *block;
  
  while(cache_current_size + size > MAX_CACHE_SIZE)
    remove_LRU_cache_block();

  while(!(block = kmalloc(size, GFP_KERNEL)) && b)
    {
      printk("can\'t kmalloc, removing LRU blocks...");
      b = remove_LRU_cache_block();
    }
  if (!b) 
    printk("cache_alloc ERROR: can't allocate memory. The cache seems empty.\n");
  else 
    cache_current_size += size;

  return block;
}

